@@ -1,4 +1,5 @@
 use super::{is_zip, Error};
+use core::num::ParseIntError;
 use fxhash::{FxHashMap, FxHashSet};
 use indicatif::{ProgressBar, ProgressStyle};
 use prettytable::Table;
@@ -32,7 +33,7 @@ impl GameDb {
     }
 
     #[inline]
-    pub fn games_iter(&self) -> impl Iterator<Item = &Game> {
+    pub fn games_iter(&self) -> impl ExactSizeIterator<Item = &Game> {
         self.games.values()
     }
 
@@ -321,14 +322,15 @@ impl Game {
         let files_on_disk: DashMap<String, PathBuf> = DashMap::new();
 
         if let Ok(dir) = read_dir(game_root) {
-            for entry in dir.filter_map(|e| e.ok()) {
+            for entry in dir
+                .filter_map(|e| e.ok())
+                .filter(|e| e.file_type().map(|t| t.is_file()).unwrap_or(false))
+            {
                 match entry.file_name().into_string() {
                     Ok(name) => {
-                        if entry.file_type().map(|t| t.is_file()).unwrap_or(false) {
-                            files_on_disk.insert(name, entry.path());
-                        }
+                        files_on_disk.insert(name, entry.path()).map(|_| ());
                     }
-                    Err(_) => failures.push(VerifyFailure::Extra(entry.path())),
+                    Err(_) => failures.push(VerifyFailure::extra(entry.path())),
                 }
             }
         } else if self.parts.is_empty() {
@@ -342,8 +344,11 @@ impl Game {
             self.parts
                 .par_iter()
                 .filter_map(|(name, part)| match files_on_disk.remove(name) {
-                    Some((_, pathbuf)) => part.verify(pathbuf),
-                    None => Some(VerifyFailure::Missing(game_root.join(name))),
+                    Some((_, pathbuf)) => part.verify(pathbuf).err(),
+                    None => Some(VerifyFailure::Missing {
+                        path: game_root.join(name),
+                        part: part.clone(),
+                    }),
                 })
                 .collect::<Vec<VerifyFailure<PathBuf>>>()
                 .into_iter(),
@@ -353,22 +358,123 @@ impl Game {
         failures.extend(
             files_on_disk
                 .into_iter()
-                .map(|(_, pathbuf)| VerifyFailure::Extra(pathbuf)),
+                .map(|(_, pathbuf)| VerifyFailure::extra(pathbuf)),
         );
 
         failures
     }
 
-    pub fn add(
+    pub fn add_and_verify(
         &self,
         rom_sources: &mut RomSources,
-        target: &Path,
-        extract: ExtractFn,
-    ) -> Result<(), Error> {
-        let target_dir = target.join(&self.name);
-        self.parts
-            .iter()
-            .try_for_each(|(rom_name, part)| extract(rom_sources, part, target_dir.join(rom_name)))
+        target_dir: &Path,
+        progress: &ProgressBar,
+    ) -> Result<Vec<VerifyFailure<PathBuf>>, Error> {
+        use std::fs::read_dir;
+
+        let mut failures = Vec::new();
+
+        let game_root = target_dir.join(&self.name);
+
+        // turn files on disk into a map, so extra files can be located
+        let mut files_on_disk: HashMap<String, PathBuf> = HashMap::new();
+
+        if let Ok(dir) = read_dir(&game_root) {
+            for entry in dir
+                .filter_map(|e| e.ok())
+                .filter(|e| e.file_type().map(|t| t.is_file()).unwrap_or(false))
+            {
+                match entry.file_name().into_string() {
+                    Ok(name) => {
+                        files_on_disk.insert(name, entry.path());
+                    }
+                    Err(_) => failures.push(VerifyFailure::extra(entry.path())),
+                }
+            }
+        }
+
+        // verify all game parts
+        for (name, part) in self.parts.iter() {
+            use std::collections::hash_map::Entry;
+            use std::fs::remove_file;
+
+            match files_on_disk.remove(name) {
+                Some(target) => {
+                    // if file exists on disk
+                    if let Err(failure) = part.verify(&target).map_err(|err| err.with_path(())) {
+                        // but is not correct
+                        match rom_sources.entry(part.clone()) {
+                            Entry::Occupied(mut entry) => {
+                                // if part exists in rom_sources
+                                // replace incorrect file with file from rom_sources
+                                let source = entry.get();
+                                remove_file(&target).map_err(Error::IO)?;
+                                match source.extract(&target)? {
+                                    Extracted::Copied => {
+                                        progress.println(format!(
+                                            "{} => {}",
+                                            source,
+                                            target.display()
+                                        ));
+                                        entry.insert(RomSource::Disk(target));
+                                    }
+                                    Extracted::Linked => progress.println(format!(
+                                        "{} -> {}",
+                                        source,
+                                        target.display()
+                                    )),
+                                }
+                            }
+                            Entry::Vacant(_) => {
+                                // if part missing in rom_sources, forward error
+                                failures.push(failure.with_path(target));
+                            }
+                        }
+                    }
+                }
+                None => {
+                    // if file does not already exist on disk
+
+                    use std::fs::create_dir_all;
+
+                    let target = game_root.join(name);
+
+                    match rom_sources.entry(part.clone()) {
+                        Entry::Occupied(mut entry) => {
+                            // and if part exists in rom_sources
+                            // link/copy file from rom_sources
+                            let source = entry.get();
+                            create_dir_all(target.parent().unwrap())?;
+                            match source.extract(&target)? {
+                                Extracted::Copied => {
+                                    progress.println(format!("{} => {}", source, target.display()));
+                                    entry.insert(RomSource::Disk(target));
+                                }
+                                Extracted::Linked => {
+                                    progress.println(format!("{} -> {}", source, target.display()))
+                                }
+                            }
+                        }
+                        Entry::Vacant(_) => {
+                            // otherwise mark as missing
+                            failures.push(VerifyFailure::Missing {
+                                path: target,
+                                part: part.clone(),
+                            });
+                        }
+                    }
+                }
+            }
+        }
+
+        // mark any leftover files on disk as extras
+        failures.extend(
+            files_on_disk
+                .into_iter()
+                .map(|(_, pathbuf)| VerifyFailure::extra(pathbuf)),
+        );
+
+        Ok(cleanup_failures(failures, progress))
     }
 
     pub fn rename(
@@ -450,20 +556,65 @@ impl<'a> GameRow<'a> {
 }
 
 pub enum VerifyFailure<P> {
-    Missing(P),
-    Extra(P),
-    Bad(P),
-    Error(P, std::io::Error),
+    Missing {
+        path: P,
+        part: Part,
+    },
+    Extra {
+        path: P,
+        part: Result<Part, std::io::Error>,
+    },
+    Bad {
+        path: P,
+        expected: Part,
+        actual: Part,
+    },
+    Error {
+        path: P,
+        err: std::io::Error,
+    },
+}
+
+impl<P: AsRef<Path>> VerifyFailure<P> {
+    #[inline]
+    fn extra(path: P) -> Self {
+        Self::Extra {
+            part: Part::from_path(path.as_ref()),
+            path,
+        }
+    }
+}
+
+impl<P> VerifyFailure<P> {
+    #[inline]
+    fn with_path<Q>(self, path: Q) -> VerifyFailure<Q> {
+        match self {
+            VerifyFailure::Missing { part, path: _ } => VerifyFailure::Missing { path, part },
+            VerifyFailure::Extra { part, path: _ } => VerifyFailure::Extra { path, part },
+            VerifyFailure::Bad {
+                expected,
+                actual,
+                path: _,
+            } => VerifyFailure::Bad {
+                path,
+                expected,
+                actual,
+            },
+            VerifyFailure::Error { err, path: _ } => VerifyFailure::Error { path, err },
+        }
+    }
 }
 
 impl<P: AsRef<Path>> fmt::Display for VerifyFailure<P> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
-            VerifyFailure::Missing(pb) => write!(f, "MISSING : {}", pb.as_ref().display()),
-            VerifyFailure::Extra(pb) => write!(f, "EXTRA : {}", pb.as_ref().display()),
-            VerifyFailure::Bad(pb) => write!(f, "BAD : {}", pb.as_ref().display()),
-            VerifyFailure::Error(pb, err) => {
-                write!(f, "ERROR : {} : {}", pb.as_ref().display(), err)
+            VerifyFailure::Missing { path, .. } => {
+                write!(f, "MISSING : {}", path.as_ref().display())
+            }
+            VerifyFailure::Extra { path, .. } => write!(f, "EXTRA : {}", path.as_ref().display()),
+            VerifyFailure::Bad { path, .. } => write!(f, "BAD : {}", path.as_ref().display()),
+            VerifyFailure::Error { path, err } => {
+                write!(f, "ERROR : {} : {}", path.as_ref().display(), err)
             }
         }
     }
@@ -524,7 +675,7 @@ impl Part {
     #[inline]
     pub fn digest(&self) -> Digest {
         match self {
-            Part::Rom { sha1 } => Digest(sha1),
+            Part::Rom { sha1, .. } => Digest(sha1),
             Part::Disk { sha1 } => Digest(sha1),
         }
     }
@@ -552,7 +703,7 @@ impl Part {
         // while generating the Part from path
         // which locks out other threads until finished
         // whereas a get()/insert() pair does not
-        let map = PART_CACHE.get_or_init(|| DashMap::default());
+        let map = PART_CACHE.get_or_init(DashMap::default);
 
         match map.get(&file_id) {
             Some(part) => Ok(part.clone()),
@@ -570,7 +721,7 @@ impl Part {
         let mut r = Sha1Reader::new(r);
         match Part::disk_from_reader(&mut r) {
             Ok(Some(part)) => Ok(part),
-            Ok(None) => copy(&mut r, &mut sink()).map(|_| Part::Rom { sha1: r.digest() }),
+            Ok(None) => copy(&mut r, &mut sink()).map(|_| r.into()),
             Err(err) => Err(err),
         }
     }
@@ -608,18 +759,20 @@ impl Part {
         Ok(Some(Part::Disk { sha1 }))
     }
 
-    fn verify<P: AsRef<Path>>(&self, part_path: P) -> Option<VerifyFailure<P>> {
+    fn verify<P: AsRef<Path>>(&self, part_path: P) -> Result<(), VerifyFailure<P>> {
         match Part::from_cached_path(part_path.as_ref()) {
-            Ok(ref disk_part) if self == disk_part => None,
-            Ok(_) => Some(VerifyFailure::Bad(part_path)),
-            Err(err) => Some(VerifyFailure::Error(part_path, err)),
+            Ok(ref disk_part) if self == disk_part => Ok(()),
+            Ok(ref disk_part) => Err(VerifyFailure::Bad {
+                path: part_path,
+                expected: self.clone(),
+                actual: disk_part.clone(),
+            }),
+            Err(err) => Err(VerifyFailure::Error {
+                path: part_path,
+                err,
+            }),
         }
     }
-
-    #[inline]
-    fn is_bad<P: AsRef<Path>>(&self, part_path: P) -> bool {
-        matches!(self.verify(part_path), Some(VerifyFailure::Bad(_)))
-    }
 }
 
 struct Sha1Reader<R> {
@@ -635,11 +788,6 @@ impl<R> Sha1Reader<R> {
             sha1: Sha1::new(),
         }
     }
-
-    #[inline]
-    fn digest(self) -> [u8; 20] {
-        self.sha1.digest().bytes()
-    }
 }
 
 impl<R: Read> Read for Sha1Reader<R> {
@@ -650,6 +798,15 @@ impl<R: Read> Read for Sha1Reader<R> {
     }
 }
 
+impl<R> From<Sha1Reader<R>> for Part {
+    #[inline]
+    fn from(other: Sha1Reader<R>) -> Part {
+        Part::Rom {
+            sha1: other.sha1.digest().bytes(),
+        }
+    }
+}
+
 pub fn parse_sha1(hex: &str) -> [u8; 20] {
     let mut hex = hex.trim();
     let mut bin = [0; 20];
@@ -912,9 +1069,9 @@ enum Extracted {
     Linked,
 }
 
-type RomSources = FxHashMap<Part, RomSource>;
+pub type RomSources = FxHashMap<Part, RomSource>;
 
-fn rom_sources<F>(root: &Path, filter: F) -> RomSources
+fn rom_sources<F>(root: &Path, part_filter: F) -> RomSources
 where
     F: Fn(&Part) -> bool + Sync + Send,
 {
@@ -935,7 +1092,7 @@ where
                 .unwrap_or_else(|_| Vec::new())
                 .into_par_iter()
         })
-        .filter(|(part, _)| filter(part))
+        .filter(|(part, _)| part_filter(part))
         .collect();
 
     pbar.finish_and_clear();
@@ -951,72 +1108,6 @@ pub fn get_rom_sources(root: &Path, required: FxHashSet<Part>) -> RomSources {
     rom_sources(root, |part| required.contains(part))
 }
 
-pub type ExtractFn = fn(&mut RomSources, &Part, PathBuf) -> Result<(), Error>;
-
-pub fn extract(roms: &mut RomSources, part: &Part, target: PathBuf) -> Result<(), Error> {
-    use std::collections::hash_map::Entry;
-
-    if !target.exists() {
-        if let Entry::Occupied(mut entry) = roms.entry(part.clone()) {
-            use std::fs::create_dir_all;
-
-            create_dir_all(target.parent().unwrap())?;
-
-            let source = entry.get();
-            match source.extract(&target)? {
-                Extracted::Copied => {
-                    println!("{} => {}", source, target.display());
-                    entry.insert(RomSource::Disk(target));
-                }
-                Extracted::Linked => println!("{} -> {}", source, target.display()),
-            }
-        }
-        Ok(())
-    } else if let Some(VerifyFailure::Bad(target)) = part.verify(target) {
-        use std::fs::remove_file;
-
-        remove_file(&target)
-            .map_err(Error::IO)
-            .and_then(|()| extract(roms, part, target))
-    } else {
-        Ok(())
-    }
-}
-
-pub fn extract_dry_run(roms: &mut RomSources, part: &Part, target: PathBuf) -> Result<(), Error> {
-    use std::collections::hash_map::Entry;
-
-    if !target.exists() || part.is_bad(&target) {
-        if let Entry::Occupied(entry) = roms.entry(part.clone()) {
-            match entry.get() {
-                RomSource::Disk(source) => {
-                    println!("{} -> {}", source.display(), target.display());
-                }
-                RomSource::Zip {
-                    file: source,
-                    index,
-                } => {
-                    println!("{}:{} -> {}", source.display(), index, target.display());
-                }
-                RomSource::SubZip {
-                    file: source,
-                    index,
-                    sub_index,
-                } => {
-                    println!(
-                        "{}:{}:{} -> {}",
-                        source.display(),
-                        index,
-                        sub_index,
-                        target.display()
-                    );
-                }
-            }
-        }
-    }
-    Ok(())
-}
-
 pub fn file_move(source: &Path, target: &Path) -> Result<(), std::io::Error> {
     if (source != target) && !target.exists() {
         use std::fs::rename;
@@ -1062,3 +1153,128 @@ pub fn display_bad_results(game: &str, failures: &[VerifyFailure<PathBuf>]) {
         }
     }
 }
+
+#[inline]
+pub fn parse_int(s: &str) -> Result<u64, ParseIntError> {
+    // MAME's use of integer values is a horror show
+    let s = s.trim();
+
+    u64::from_str(s)
+        .or_else(|_| u64::from_str_radix(s, 16))
+        .or_else(|e| {
+            if let Some(stripped) = s.strip_prefix("0x") {
+                u64::from_str_radix(stripped, 16)
+            } else {
+                dbg!(s);
+                Err(e)
+            }
+        })
+}
+
+fn cleanup_failures<P: AsRef<Path>>(
+    failures: Vec<VerifyFailure<P>>,
+    progress: &ProgressBar,
+) -> Vec<VerifyFailure<P>> {
+    if failures.is_empty() {
+        failures
+    } else {
+        let mut extras = HashMap::new();
+        let mut to_cleanup = Vec::new();
+
+        for failure in failures {
+            match failure {
+                VerifyFailure::Extra {
+                    path,
+                    part: Ok(part),
+                } => {
+                    extras.insert(part, path);
+                }
+                other => {
+                    to_cleanup.push(other);
+                }
+            }
+        }
+
+        if extras.is_empty() {
+            to_cleanup
+        } else {
+            use std::fs::rename;
+
+            let mut failures = Vec::new();
+
+            for cleanup in to_cleanup {
+                match cleanup {
+                    VerifyFailure::Missing {
+                        path: missing_path,
+                        part,
+                    } => match extras.remove(&part) {
+                        Some(extra_path) => {
+                            if rename(extra_path.as_ref(), missing_path.as_ref()).is_ok() {
+                                progress.println(format!(
+                                    "{} -> {}",
+                                    extra_path.as_ref().display(),
+                                    missing_path.as_ref().display()
+                                ));
+                            } else {
+                                failures.push(VerifyFailure::Missing {
+                                    path: missing_path,
+                                    part,
+                                });
+                            }
+                        }
+                        None => failures.push(VerifyFailure::Missing {
+                            path: missing_path,
+                            part,
+                        }),
+                    },
+                    VerifyFailure::Bad {
+                        path: bad_path,
+                        expected,
+                        actual,
+                    } => match extras.remove(&expected) {
+                        Some(extra_path) => {
+                            if std::fs::remove_file(bad_path.as_ref())
+                                .and_then(|()| rename(extra_path.as_ref(), bad_path.as_ref()))
+                                .is_ok()
+                            {
+                                progress.println(format!(
+                                    "{} -> {}",
+                                    extra_path.as_ref().display(),
+                                    bad_path.as_ref().display()
+                                ));
+                            } else {
+                                failures.push(VerifyFailure::Bad {
+                                    path: bad_path,
+                                    expected,
+                                    actual,
+                                });
+                            }
+                        }
+                        None => failures.push(VerifyFailure::Bad {
+                            path: bad_path,
+                            expected,
+                            actual,
+                        }),
+                    },
+                    error @ VerifyFailure::Error { .. } => failures.push(error),
+                    extra @ VerifyFailure::Extra { .. } => failures.push(extra),
+                }
+            }
+
+            failures.extend(extras.into_iter().filter_map(
+                |(part, path)| match std::fs::remove_file(path.as_ref()) {
+                    Ok(()) => {
+                        progress.println(format!("deleted : {}", path.as_ref().display()));
+                        None
+                    }
+                    Err(_) => Some(VerifyFailure::Extra {
+                        part: Ok(part),
+                        path,
+                    }),
+                },
+            ));
+
+            failures
+        }
+    }
+}
